""" Optional constraint functions extracted for
:func:`~despasito.parameter_fitting.fit_functions.global_minimization` method.
"""


def type1(x, *args):
    """Example constraint. Notice that if this is used by a constraint class, there
    are no other args, while there are for dict constraints."""

    output = None

    return output
